#! /usr/bin/env python
"""Create .deb sets from unattended upgrade"""
from __future__ import print_function
from fussy import nbio, twrite
import re, os, time, logging, json
import tarfile

log = logging.getLogger(__name__)
TARGET_MATCHER = re.compile(r"""DestFile:['](?P<filename>[^']+)[']""")
DEPENDENCY_MATCHER = re.compile(
    r"""(?P<package>[^, ()]*?)[ ]*[(](?P<compare>[<>=]+)[ ]*(?P<version>[^)]*?)[)]"""
)
DIGIT_FINDER = re.compile('\d+')


def parse_dpkg_info(content):
    versions = {}
    content = content.strip().splitlines()
    current = None
    for line in content:
        for interesting in [
            'Package:',
            'Version:',
            'Status:',
            'Depends:',
            'Pre-Depends:',
            'Architecture:',
        ]:
            line = line.strip()
            if line.startswith(interesting):
                value = line[len(interesting) :].strip()
                if interesting == 'Package:':
                    if current:
                        versions[current['package']] = current
                    current = dict(package=value)
                elif interesting == 'Version:':
                    version = line.split(':', 1)[1].strip()
                    current['version'] = [
                        int(x, 10) for x in DIGIT_FINDER.findall(version)
                    ]
                elif interesting == 'Status:':
                    if 'installed' in value.split():
                        current['installed'] = True
                    else:
                        current['installed'] = False
                elif interesting in ('Depends:', 'Pre-Depends:'):
                    value = [x.groupdict() for x in DEPENDENCY_MATCHER.finditer(value)]
                    if interesting == 'Depends:':
                        key = 'depends'
                    else:
                        key = 'predepends'
                    current[key] = value
                else:
                    current[interesting.lower().rstrip(':')] = value
    if current:
        versions[current['package']] = current
    return versions


def get_options():
    import argparse

    parser = argparse.ArgumentParser(
        description='Process unattended upgrade log to produce upgrade set'
    )
    default_file = 'os-upgrades-%s.deb.tar' % (
        time.strftime('%Y-%m-%d-%H-%M', time.gmtime())
    )
    parser.add_argument(
        '--target',
        metavar='FILENAME.deb.tar',
        default=default_file,
        help="Target directory to watch (must exist), default: %s" % (default_file,),
    )
    parser.add_argument(
        '--list',
        action='store_true',
        default=False,
        help="Just list the files that would be packed, do not do actual packing",
    )
    parser.add_argument(
        '--exclude',
        metavar='PACKAGE',
        dest='excluded',
        action='append',
        help='Packages *not* to include in the resulting tar-file',
    )
    parser.add_argument(
        '--recalculate',
        action='store_true',
        default=False,
        help='If provided, will run unattended upgrades with dry-run to calculate (and download) the set',
    )
    return parser


def get_metadata_options():
    import argparse

    parser = argparse.ArgumentParser(
        description='Extract metadata from .deb files for unattended set'
    )
    parser.add_argument(
        '-m',
        '--merge',
        metavar='FULLPATH.json',
        default=None,
        help='If provided, metadata will be *merged* into existing json file',
    )
    parser.add_argument('filenames', metavar='FULLPATH.deb', nargs="+")
    return parser


def recalculate():
    log.warning("Running unattended upgrades, this will take a long time")
    nbio.Process('sudo unattended-upgrades -d --dry-run')()


def filter_exclusions(package, excluded):
    import fnmatch

    for exclude in excluded:
        if fnmatch.fnmatch(package, exclude):
            log.info("Excluding %s due to pattern %s", package, exclude)
            return False
    return True


def get_components(excluded=None):
    result = []
    unique = set()
    for line in nbio.Process(
        """grep '/var/cache/apt/archives' /var/log/unattended-upgrades/unattended-upgrades.log | grep -v check_conffile | grep AcquireItem"""
    )().splitlines():
        match = TARGET_MATCHER.search(line)
        if match:
            filename = match.group('filename')
            base = os.path.basename(filename)
            # Use package name + Architecture (i.e. amd64, i386, all)
            package = '_'.join([base.split('_')[0], base.split('_')[-1]])
            if package in unique:
                continue
            unique.add(package)
            if excluded:
                if filter_exclusions(package, excluded):
                    result.append(filename)
            else:
                result.append(filename)
    return result


def metadata_main():
    options = get_metadata_options().parse_args()
    structure = metadata_structure(options.filenames)
    if options.merge:
        new_set = dict([(s['package'], s) for s in structure['metadata']])
        final_set = []
        previous = json.loads(open(options.merge).read())
        for item in previous['metadata']:
            if item['package'] in new_set:
                log.info("Updating record for %s", item['package'])
                item = new_set.pop(item['package'])
            final_set.append(item)
        final_set.extend(new_set.values())
        structure['metadata'] = final_set
        previous.update(structure)
        structure = previous
    print(json.dumps(structure, indent=2))


def metadata_structure(filenames):
    metadata_list = []
    for package in filenames:
        metadata_list.append(metadata(package))
    return {'metadata': metadata_list, 'ts': time.time()}


def metadata(filename):
    command = ['dpkg', '--info', filename]

    output = nbio.Process(command)()
    for record in parse_dpkg_info(output).values():
        record['filename'] = os.path.basename(filename)
        return record
    raise RuntimeError("Unable to extract metadata for %r" % (filename,))


def pack_zip(target='os-upgrades.deb.tar', excluded=None):
    log.info("Packing into target: %r", target)
    t = tarfile.TarFile(target, mode='w')
    try:
        packages = get_components(excluded=excluded)
        for package in packages:
            log.info("Packing: %s", package)
            t.add(package, os.path.basename(package))
        structure = metadata_structure(packages)
        try:
            twrite.twrite(target + '.json', json.dumps(structure))
            t.add(target + '.json', 'metadata.json')
        finally:
            os.remove(target + '.json')
    finally:
        t.close()


def main():
    logging.basicConfig(level=logging.INFO)
    options = get_options().parse_args()
    if options.recalculate:
        recalculate()
    if options.list:
        for package in get_components(excluded=options.excluded):
            base = os.path.basename(package)
            package, version = base.split('_', 1)
            print('%s\t%s' % (package, version))
        return
    pack_zip(target=options.target, excluded=options.excluded)
    return 0


if __name__ == "__main__":
    main()



"""Transactionally write a file to disk"""
import os, logging, shutil
from ._shims import bytes, unicode

log = logging.getLogger(__name__)


def twrite(filename, content, ownership=None, mode=None):
    """Write content to filename, creating directory if it does not exist

    ownership -- None => no change
                 (uid,gid) => if either is None, use euid(),egid()
                 otherwise set to the value specified
    mode -- None => no change
            -1 => copy from the target
            other => passed to os.chmod()
    Notes:
        Is not race-condition safe wrt multiple writers, as it uses 
        the same filename for the temp file in each write.
        
        Transactionality is Linux specific (os.rename guarantee)
    """
    directory = os.path.dirname(filename)
    if directory and not os.path.exists(directory):
        log.debug('Creating directory: %s', directory)
        os.makedirs(directory)
    tmp = os.path.join(filename + '~')
    fh = open(tmp, 'wb')
    if isinstance(content, bytes):
        fh.write(content)
    elif isinstance(content, unicode):
        fh.write(content.encode('utf-8'))
    else:
        for item in content:
            fh.write(item)
    fh.close()
    if ownership:
        (uid,gid) = ownership 
        if uid is None:
            uid = os.geteuid()
        if gid is None:
            gid = os.getegid()
        os.chown(tmp,uid,gid)
    if mode is None:
        pass 
    elif mode == -1:
        if os.path.exists(filename):
            shutil.copymode(filename, tmp)
    else:
        os.chmod(tmp,mode)
    log.debug('Writing file: %s', filename)
    try:
        os.rename(tmp, filename)
    except OSError as err:
        if err.errno == 2:  # no such directory
            pass
        else:
            raise



"""Provide a cron lock for preventing cron-bombs and the like"""
import os, logging, fcntl, errno, signal, tempfile, time, threading

log = logging.getLogger(__name__)


def time_function(marker=None):
    from functools import wraps

    def wrapper(function):
        if marker is None:
            final_marker = '%s.%s' % (function.__module__, function.__name__)
        else:
            final_marker = marker

        @wraps(function)
        def timer(*args, **named):
            start = time.time()
            try:
                holder = open('/var/firmware/run/request-localhost-8111.lock').read()
            except Exception:
                holder = None
            try:
                return function(*args, **named)
            finally:
                log.info("%s: %0.1fs by %s", final_marker, time.time() - start, holder)

        return timer

    return wrapper


class Busy(IOError):
    """Raised if the lock is held by another cronlock"""


class Timeout(RuntimeError):
    """Raised if the timeout signal triggers"""


class AlreadyHeld(Busy):
    """Raised if we attempt to lock multiple times from the same process"""


class Flock(object):
    """A context manager that just flocks a file"""

    file = None

    def __init__(self, filename, blocking=False):
        self.filename = filename
        self.blocking = blocking

    def __str__(self):
        return '%s( %r )' % (self.__class__.__name__, self.filename)

    def __enter__(self):
        if self.file:
            raise Busy("""Attempted to lock the %s twice""" % (self))
        fh = open(self.filename, 'ab+')
        for i in range(2):  # limited number of retries
            try:
                flags = fcntl.LOCK_EX
                if not self.blocking:
                    flags |= fcntl.LOCK_NB
                fcntl.flock(fh, flags)  # can raise IOError
            except IOError as err:
                if err.errno == errno.EWOULDBLOCK:
                    other_pid = None
                    if os.path.exists(self.filename):
                        try:
                            other_pid = int(open(self.filename).readline())
                        except Exception:
                            pass
                        else:
                            from fussy import processcontrol

                            if os.getpid() == other_pid:
                                raise AlreadyHeld(
                                    "Multiple attempts to lock the same file from the same process %s"
                                    % (self.filename)
                                )
                            if not processcontrol.alive(other_pid):
                                log.info(
                                    "PID#%s held the lock on %s, but does not appear to be alive, retrying",
                                    other_pid,
                                    self.filename,
                                )
                                continue

                    err = Busy(*(err.args + ('held by PID#%s' % (other_pid,),)))
                    err.errno = errno.EWOULDBLOCK
                raise err
        fh.seek(0)  # rewind to start...
        fh.write((u'%s\n' % (os.getpid())).encode('utf-8'))
        fh.flush()
        self.file = fh
        return fh

    def __exit__(self, *args):
        try:
            fh = self.__dict__.pop('file')
        except KeyError:
            log.warning(
                "No 'file' member in lock, somehow exiting before entry completed?"
            )
        else:
            try:
                os.remove(self.filename)
            except (OSError, IOError):
                log.warning('File %s removed while locked', self.filename)
            fcntl.flock(fh, fcntl.LOCK_UN)
            fh.close()

    def write(self, *args):
        with self:
            return self._write(*args)

    def _write(self, *args):
        self.file.seek(0)
        self.file.write(*args)
        self.file.flush()

    def read(self, *args):
        with self:
            return self._read(*args)

    def _read(self, *args):
        return self.file.read(*args)


class Lock(object):
    """A Context manager that provides cron-style locking"""

    def __init__(self, lockfile):
        """Create a lock file 
        
        name -- used to construct the lock-file name 
        directory -- directory in which to construct the lock-file 
        """
        self.flock = Flock(lockfile)

    @property
    def pid(self):
        return os.getpid()

    def __enter__(self):
        self.flock.__enter__()

    def __exit__(self, *args):
        self.flock.__exit__(*args)

    def on_timeout(self, *args, **named):
        raise Timeout("Maximum run-time exceeded, aborting")

    def set_timeout(self, duration):
        """Set a signal to fire after duration and raise an error"""
        signal.signal(signal.SIGALRM, self.on_timeout)
        signal.alarm(int(duration))


class ThreadLock(object):
    """Simple context manager for a threading RLock"""

    def __init__(self, lock_wait=5):
        self._lock = threading.RLock()
        self.lock_wait = lock_wait

    def __enter__(self):
        start = time.time()
        abort = start + self.lock_wait
        while time.time() < abort:
            if not self._lock.acquire(blocking=False):
                time.sleep(max(((abort - time.time()) / 100.0, 0.00001)))
            else:
                # log.debug("Acquired thread lock after: %0.3fs", time.time()-start)
                return self
        raise Busy('Timeout waiting for thread lock')

    def __exit__(self, *args):
        self._lock.release()


class BlockingLock(Lock):
    """Lock class that blocks for up to lock_wait seconds waiting for other process to exit
    
    Also uses ThreadLock to prevent multiple threads from 
    this process from running simultaneously...
    """

    def __init__(self, lockfile, lock_wait=5):
        super(BlockingLock, self).__init__(lockfile)
        self._threadlock = ThreadLock()
        self.lock_wait = lock_wait

    def __enter__(self):
        start = time.time()
        abort = start + self.lock_wait
        while time.time() < abort:
            try:
                result = super(BlockingLock, self).__enter__()
                delta = time.time() - start
                if delta > 0.1:
                    log.info("Acquired after %0.1fs", delta)
                return result
            except Busy:
                max_wait = 0.01
                duration = max(
                    (min(((abort - time.time()) / 100.0, max_wait)), 0.00001)
                )
                time.sleep(duration)
        log.debug("Timeout waiting")
        raise Busy("Lock on %s timed out" % (self.flock.filename))


def with_lock(name, directory=None, timeout=None):
    """Decorator that runs a function with Lock instance acquired
    
    * name -- basename of the file to create 
    * directory -- if specified, the directory in which to store files, 
        defaults to tempfile.gettempdir()
    * timeout -- if specified, the number of seconds to allow before raising 
        a Timeout error
    """
    if directory is None:
        directory = tempfile.gettempdir()
    filename = os.path.join(directory, name)
    lock = Lock(filename)

    def wrap_func(function):
        """Wraps the function execution with our lock"""

        def wrapped_with_lock(*args, **named):
            with lock:
                if timeout:
                    lock.set_timeout(timeout)
                return function(*args, **named)

        wrapped_with_lock.__name__ = function.__name__
        wrapped_with_lock.__doc__ = function.__doc__
        wrapped_with_lock.__dict__ = function.__dict__
        wrapped_with_lock.lock = lock
        return wrapped_with_lock

    return wrap_func



"""Tail protocol for files using Twisted non-blocking IO"""
import os, sys, errno
from twisted.internet import fdesc, reactor
from twisted.python import log

from twisted.internet import inotify
from twisted.python import filepath


class FileTail(object):
    def __init__(self, filename):
        if hasattr(filename, 'path'):
            filename = filename.path
        self.filename = filename
        self.buffer = ''

    def start_watching(self):
        # TODO: allow for handling log rotation etc.
        # TODO: use the overall directory inotify instead,
        # and use *that* to do log-rotation checks...
        self.notifier = inotify.INotify()
        self.notifier.startReading()
        self.notifier.watch(
            filepath.FilePath(self.filename),
            mask=inotify.IN_MODIFY,
            callbacks=[self.on_file_change],
        )

    _fd = None

    @property
    def fd(self):
        if self._fd is None:
            self._fd = os.open(self.filename, os.O_RDONLY)
            fdesc.setNonBlocking(self._fd)
        return self._fd

    def on_file_change(self, ignored, filepath, mask):
        if mask & inotify.IN_MODIFY:
            self.read_from_file()

    def read_from_file(self):
        fd = self.fd
        while True:
            # TODO: really, we should do a non-blocking producer
            # using an abstract file handle, but as that turns
            # out not to usefully restrict us on EOF it's kinda
            # loopy...
            # For the simple "tail a file to web browser" case
            # this should work...
            try:
                output = os.read(fd, 8192)
            except (OSError, IOError) as err:
                if err.args[0] in (errno.EAGAIN, errno.EINTR):
                    pass
                elif err.args[0] in errno.EWOULDBLOCK:
                    # can't read any more
                    break
                else:
                    log.msg("Error reading from %s" % (self.name,))
                    self._fd = None
            else:
                if output:
                    self.buffer += output
                    log.msg('Got output: %s' % (len(self.buffer)))
                else:
                    # reached end of file...
                    break
        if self.buffer:
            lines = self.buffer.splitlines()
            if not self.buffer.endswith('\n'):
                self.buffer = lines.pop()
            else:
                self.buffer = ''
            return lines
        return []


def test_consumer(lines):
    for line in lines:
        log.msg('line: %r' % (line,))


def main():
    log.startLogging(sys.stderr)
    FileTail(sys.argv[1], test_consumer).start_watching()
    reactor.run()


if __name__ == "__main__":
    main()



"""Load firmware configuration so that we can find out firmware metadata

Expected keys:

    [fussy]
    product=sample
    keyring=/etc/fussy/keys
    extra_etc=/etc/%(product)s
    var_directory=/var/%(sample)s
    firmware_directory=/opt/%(sample)s
"""
import glob

try:
    import ConfigParser as configparser
except ImportError:
    import configparser

options = None


def loadoptions(product='fussy'):
    global options
    if options is None:
        options = configparser.SafeConfigParser()
        options.read(sorted(glob.glob('/etc/fussy/*.conf')))
        try:
            product = options.get('fussy', 'product')
        except (configparser.NoSectionError, configparser.NoOptionError) as err:
            pass
        else:
            options.read(sorted(glob.glob('/etc/%(product)s/*.conf' % locals())))
    return options


def getstring(section, key, default=None):
    if not options:
        raise RuntimeError("Need to load options first")
    try:
        return options.get(section, key)
    except (configparser.NoSectionError, configparser.NoOptionError):
        return default


def getboolean(section, key, default=None):
    if not options:
        raise RuntimeError("Need to load options first")
    try:
        return options.getboolean(section, key)
    except (configparser.NoSectionError, configparser.NoOptionError):
        return default



#! /usr/bin/env python
import json, sys


def main():
    if sys.argv[1:]:
        source = open(sys.argv[1]).read()
    else:
        source = sys.stdin.read()
    content = json.loads(source)
    sys.stdout.write(json.dumps(content, indent=2, sort_keys=True))


if __name__ == "__main__":
    main()



"""Link all files in a distribution from relative location on the hard disk"""
import os, logging, shutil
from optparse import OptionParser

log = logging.getLogger(__name__)


def link_tree(distribution_dir, target_dir='/', symbolic=True, copy=False):
    """(Sym)link all files under distribution_dir from target_dir 
    """
    if symbolic:
        link = os.symlink
    else:
        link = os.link
    for path, subdirs, filenames in os.walk(distribution_dir):
        dir_relative = os.path.relpath(path, distribution_dir)
        dir_source = os.path.normpath(os.path.abspath(path))
        dir_final = os.path.normpath(os.path.join(target_dir, dir_relative))
        if not os.path.exists(dir_final):
            log.info('Creating directory: %s', dir_final)
            try:
                os.makedirs(dir_final)
            except Exception as err:
                err.args += (dir_final,)
                raise
        for filename in filenames + subdirs:
            file_source = os.path.join(dir_source, filename)
            file_final = os.path.join(dir_final, filename)
            if os.path.islink(file_source):

                target = os.readlink(file_source)
                final_target = os.path.join(dir_final, target)
                log.info('Recreating link in %s to %s', file_final, final_target)
                create_link(final_target, file_final, os.symlink)
            elif os.path.isfile(file_source):
                if os.path.lexists(file_final):
                    if (
                        symbolic
                        and os.path.islink(file_final)
                        and os.readlink(file_final) == file_source
                    ):
                        log.info('Up to date: %s', file_final)
                        continue
                    log.info('Replacing: %s', file_final)
                    create_link(file_source, file_final, link, copy=copy)
                else:
                    log.info('Linking: %s', file_final)
                    create_link(file_source, file_final, link, copy=copy)


def create_link(file_source, file_final, link, copy=False):
    try:
        temp = file_final + '~'
        if os.path.exists(temp) or os.path.lexists(temp):
            try:
                os.remove(temp)
            except Exception as err:
                err.args += (temp,)
                raise
        try:
            if copy:
                shutil.copy2(file_source, temp)
            else:
                link(file_source, temp)
        except OSError as err:
            if link is os.link:
                shutil.copy2(file_source, temp)
            else:
                raise
        os.rename(temp, file_final)
    except Exception as err:
        log.error("Error linking %s: %s", file_final, err)
        raise err


def get_options():
    """Creates the OptionParser used in :func:`main` """
    parser = OptionParser()
    parser.add_option(
        '-d',
        '--directory',
        dest='directory',
        default=None,
        help="Directory to which to link",
    )
    parser.add_option(
        '-t',
        '--target',
        dest='target',
        default=None,
        help="Directory in which to create the links",
    )
    parser.add_option(
        '-v',
        '--verbose',
        dest='verbose',
        action="store_true",
        default=False,
        help="Perform verbose logging (INFO level), default: False",
    )
    parser.add_option(
        '-H',
        '--hard',
        dest='hardlink',
        action="store_true",
        default=False,
        help="Use hard-links instead of symlinks, default: False",
    )
    parser.add_option(
        '-c',
        '--copy',
        dest='copy',
        action="store_true",
        default=False,
        help="Copy the inode rather than linking, default: False",
    )
    return parser


def main():
    options, args = get_options().parse_args()
    if options.verbose:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)
    if args:
        if not options.directory:
            options.directory = args[0]
            args = args[1:]
        if args and not options.target:
            options.target = args[0]
            args = args[1:]
    link_tree(
        options.directory,
        options.target,
        symbolic=(not options.hardlink),
        copy=options.copy,
    )
    return 0



"""Utility functions for common tasks"""
from fussy import install, nbio
import os, glob, logging
from optparse import OptionParser

log = logging.getLogger(__name__)

MIGRATION_PATTERN = '????_*.py'


def find_reverse_upgrades(
    final_target, relative_path, migration_pattern=MIGRATION_PATTERN
):
    """Find reverse upgrades that must be run before final_target is installed
    
    final_target -- final target of the installation
    relative_path -- path from final target to Django/South migrations
    
    returns ID/number of the migration to run (if any)
    """
    new = glob.glob(os.path.join(final_target, relative_path, migration_pattern))
    old = glob.glob(
        os.path.join(
            final_target, '..', install.CURRENT_LINK, relative_path, migration_pattern
        )
    )
    new = sorted([os.path.basename(x) for x in new])
    old = sorted([os.path.basename(x) for x in old])
    # we want the last migration in both old and new
    common = [
        old_item for (new_item, old_item) in zip(new, old) if new_item == old_item
    ]
    if common == old:
        # forward migration only
        log.info('No reverse migration required')
        return None
    elif common:
        # there is a common set, but it is *not* the same as old, so roll back current
        # until it is at the common branching point...
        log.warning(
            "Extra migrations in current versus target: %s",
            ", ".join(old[len(common) :]),
        )
        return os.path.splitext(common[-1])[0]
    else:
        # I can't think of a legitimate use case for wanting to roll back to having no data...
        raise RuntimeError("No common migrations found")


def get_options():
    """Creates the OptionParser used in :func:`main` """
    parser = OptionParser()
    parser.add_option(
        '-f',
        '--final_target',
        dest='final_target',
        help="Path to the new firmware installation",
    )
    parser.add_option(
        '-d',
        '--django-admin',
        dest='django_admin',
        default='django-admin.py',
        help="Path to the django-admin.py script with which to run migrations",
    )
    parser.add_option(
        '-p',
        '--path',
        dest='path',
        help="Relative path from current to migration directory...",
    )
    parser.add_option('-a', '--app', dest='app', help="Application name to migrate")
    return parser


def main():
    """Finds last common migration and reverts current Django db to that 
    
    Must be run with DJANGO_SETTINGS_MODULE set in the environment
    """
    logging.basicConfig(level=logging.INFO)
    options, args = get_options().parse_args()
    migration = find_reverse_upgrades(options.final_target, options.path)
    if migration:
        pipe = nbio.Process(
            [options.django_admin, 'migrate', '--noinput', options.app, migration],
            stderr=-1,
        )
        print(pipe())



"""Unpack a signed .tar.gz.asc or .tar.gz.gpg archive into a temporary directory
"""
import tempfile, os, shutil, glob, logging, traceback
from optparse import OptionParser
from fussy import errors, nbio

log = logging.getLogger(__name__)

DEFAULT_KEYRING = '/etc/fussy/keys'


def verify_requirements(directory, requirements):
    """Verify that the directory has the expected files
    
    requirements -- list of files expected to exist in the directory,
        should be minimal, we've already crypto-verified the source, 
        this is just to prevent signed-but-defective installations
    
    raises 
    """
    for requirement in requirements:
        if not os.path.exists(os.path.join(directory, requirement)):
            raise RuntimeError("Missing path: %r" % (requirement,))


def unpack(filename, keyring=DEFAULT_KEYRING):
    """Unpack (uploaded) filename into temporary directory
    """
    source = os.path.abspath(os.path.normpath(filename))
    directory = tempfile.mkdtemp(prefix='fussy-', suffix='-unpacked')
    try:
        env = os.environ.copy()
        env['GNUPGHOME'] = keyring
        try:
            temptar = os.path.join(directory, 'firmware.tar.gz')
            nbio.Process(['gpg', '-o', temptar, '-d', source], env=env, stderr=False)()
            nbio.Process(['tar', '-zxf', temptar], cwd=directory, stderr=False)()
        except nbio.ProcessError as err:
            # log.error( "Traceback: %s", traceback.format_exc())
            try:
                output = '\n'.join(getattr(err.process, 'output', []))
            except Exception as err:
                output = ''
            if err.process.command[0] == 'gpg':
                log.error('gpg reported failure: %s', output)
                raise errors.SignatureFailure("Unable to verify signature")
            else:
                log.error('tar reported failure: %s', output)
                raise errors.ArchiveFailure(
                    "Unable to unpack the archive: %s" % (str(err))
                )
        except Exception as err:
            log.error("Unhandled error Traceback: %s", traceback.format_exc())
            raise
        else:
            os.remove(temptar)
        # there should now be *precisely* one firmware directory and
        # potentially 2 upgrade scripts (which don't match *)
        files = glob.glob(os.path.join(directory, '*'))
        if len(files) != 1:
            raise errors.ArchiveFailure(
                "Expected a single root directory, got: %s" % (files,)
            )
        directory = files[0]
        if os.path.basename(directory) in ['current']:
            raise errors.ArchiveFailure(
                "Do not allow creation of archives with name current"
            )
    except Exception as err:
        shutil.rmtree(directory, True)
        raise
    return directory


def get_options():
    parser = OptionParser()
    parser.add_option(
        '-f',
        '--file',
        dest='file',
        default=None,
        action="store",
        type="string",
        help="The firmware archive to unpack, must be a .tar.gz.gpg or a .tar.gz.asc",
    )
    parser.add_option(
        '-k',
        '--keyring',
        dest='keyring',
        default='/etc/fussy/keys',
        action="store",
        type="string",
        help="GPG keyring to use for verification/decryption (default /etc/fussy/keys)",
    )
    parser.add_option(
        '-v',
        '--verbose',
        dest='verbose',
        action='store_true',
        default=False,
        help='Do verbose logging',
    )
    return parser


def main():
    parser = get_options()
    options, args = parser.parse_args()
    logging.basicConfig(level=[logging.INFO, logging.DEBUG][bool(options.verbose)])
    if not options.file:
        if args:
            options.file = args[0]
        else:
            parser.error("Need a file to unpack")
    filename = options.file or args[0]
    log.info('Unpacking: %s', filename)
    log.info('Keyring: %s', options.keyring)
    directory = unpack(filename, keyring=options.keyring)
    print(directory)
    return 0



"""Errors raise by Fussy explicitly
"""


class FussyError(Exception):
    """Base class on which all other Fussy specific errors are based"""

    EXIT_CODE = 2


class SystemSetupError(RuntimeError, FussyError):
    """System configuration appears incorrect (missing required packages, etc)
    
    Recovery will likely require system-level intervention
    """

    EXIT_CODE = 40


class RevertedFailure(FussyError):
    """Error when failure occurred, but were able to revert to previous/failsafe
    
    No recovery is required, the previous version of the firmware was restored
    """

    EXIT_CODE = 3


class UnrecoverableError(FussyError):
    """Error raised when there's just no way to recover"""

    EXIT_CODE = 4


class SignatureFailure(FussyError):
    """Signature was incorrect, or could not be verified
    
    No recovery is required, the archive was rejected before install
    """

    EXIT_CODE = 20


class ArchiveFailure(FussyError):
    """Archive appears incorrectly structured
    
    No recovery is required, the archive was rejected before install
    """

    EXIT_CODE = 21



"""JSON log server (log-line server)

Logs each incoming log line to the disk *without modification*, but 
only writes lines when the line is complete, so allows you to log from 
many processes simultaneously.

See :py:class:`fussy.jsonlog.JSONSocketHandler` for how to send messages 
to the log server.

The log server should be set up to run under a process manager 
(e.g. supervisord) like so::

    [program:fussylog]
    command=/opt/firmware/current/env/bin/fussy-log-server -f /var/www/logs/errors.json
    autorestart=true
    user=<someuser>
    directory=/var/www/logs
    stopasgroup=true
    killasgroup=true 
    startretries=200
    priority=50
"""
import asyncore
import socket
import json
import os
import sys
import time
import logging, traceback

log = logging.getLogger(__name__)
from optparse import OptionParser
LOGGING_PORT = 2514

try:
    unicode
except NameError:
    unicode = str


class LogHandler(asyncore.dispatcher):
    """Handle messages from a single client process"""

    buffer = b''

    def __init__(self, *args, **named):
        self.sink = named.pop('sink')
        asyncore.dispatcher.__init__(self, *args, **named)

    def handle_read(self):
        try:
            update = self.recv(8192)
            if isinstance(update, unicode):
                # this should *never* happen, but it's been seen somehow...
                update = update.encode('utf-8')
            self.buffer += update
            i = 0
            while b'\n' in self.buffer and i < 50:
                index = self.buffer.index(b'\n')
                record, self.buffer = self.buffer[: index + 1], self.buffer[index + 1 :]
                if record == b'\000\n':
                    self.sink.rotate()
                else:
                    self.sink(record)
                i += 1
            if i >= 50:
                log.warning("More than 50 records received at the same time!")
        except Exception:
            log.error("Unhandled error on read: %s", traceback.format_exc())
            self.close()
            raise

    def handle_close(self):
        self.close()

    def writable(self):
        return False


class LogServer(asyncore.dispatcher):
    """Handle incoming connections from however many processes wish to log"""

    def __init__(self, host='127.0.0.1', port=LOGGING_PORT, sink=None):
        asyncore.dispatcher.__init__(self)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.set_reuse_addr()
        self.sink = sink
        self.bind((host, port))
        self.listen(5)

    def handle_accept(self):
        try:
            pair = self.accept()
            if pair is None:
                pass
            else:
                sock, addr = pair
                log.info("New connection from: %s", addr)
                LogHandler(sock, sink=self.sink)
        except Exception:
            log.error("Unhandled error on accept: %s", traceback.format_exc())
            raise


ROTATION = 1024 * 1024
ROTATION_COUNT = 5


class Sink(object):
    """Location to which to send messages (rotating files in the simple case)
    """

    file_handle = None

    def __init__(self, filename, max_bytes=ROTATION, max_rotation=ROTATION_COUNT):
        """Create a sink for the given filename"""
        self.filename = filename
        self.max_bytes = max_bytes
        self.max_rotation = max_rotation
        if os.path.exists(self.filename):
            self.written = self.last_flush = os.stat(self.filename).st_size
            self.file_handle = open(self.filename, 'ab')
        else:
            self.written = 0
            self.last_flush = 0
            self.file_handle = open(self.filename, 'wb')
        self.file_handle.write(b'\n')
        self.written += 1
        # self.rotate()

    def rotate(self):
        log.info('Rotating: %s', self.filename)
        if self.file_handle:
            self.file_handle.close()
        self.written = 0
        self.last_flush = 0
        if os.path.exists(self.filename):
            for i in range(self.max_rotation - 1, 0, -1):  # stop at 1
                rot = '%s.%s' % (self.filename, i)
                rotpone = '%s.%s' % (self.filename, i + 1)
                if os.path.exists(rot) and os.stat(rot).st_size > 0:
                    os.rename(rot, rotpone)
            i = 1
            rot = '%s.%s' % (self.filename, i)
            os.rename(self.filename, rot)
        self.file_handle = open(self.filename, 'wb')

    # By default, flush on every write, not good on a laptop, but we're logging
    # high-priority messages with this thing...
    FLUSH_SIZE = 0

    def __call__(self, record):
        self.written += len(record)
        if self.written > self.max_bytes:
            self.rotate()
            self.written = len(record)
        self.file_handle.write(record)
        if self.written - self.last_flush > self.FLUSH_SIZE:
            self.file_handle.flush()
            self.last_flush = self.written


LOGGING_PORT = 2514


def server_options():
    parser = OptionParser()
    parser.add_option(
        '-f',
        '--file',
        dest='filename',
        default='/tmp/json.log',
        help="File into which to log",
    )
    parser.add_option(
        '-r',
        '--rotation',
        dest='rotation',
        default=ROTATION,
        help="Size after which to rotate log file (bytes), defaults %s" % (ROTATION,),
    )
    parser.add_option(
        '-c',
        '--count',
        dest='rotation_count',
        default=ROTATION_COUNT,
        help="Number of rotated logs to retain, default %s" % (ROTATION_COUNT,),
    )
    parser.add_option(
        '-p',
        '--port',
        dest='port',
        default=2514,
        type='int',
        help='Port on which to listen',
    )
    return parser


def main():
    """Sink json-formatted messages into a rotating file handle"""
    parser = server_options()
    options, args = parser.parse_args()
    directory = os.path.dirname(options.filename)
    if not os.path.exists(directory):
        os.makedirs(directory)
    logging.basicConfig(
        filename=options.filename + '.logserver.log', level=logging.INFO
    )
    sink = Sink(
        filename=options.filename,
        max_bytes=options.rotation,
        max_rotation=options.rotation_count,
    )
    log.info(
        'Listening on port %s and writing to file %s', options.port, options.filename
    )
    LogServer(sink=sink, port=options.port)
    try:
        asyncore.loop()
    except KeyboardInterrupt:
        return 0

def rotate_main():
    """Do a rotate of the local log server"""
    import socket
    sock = socket.socket(socket.AF_INET)
    sock.connect(('127.0.0.1',LOGGING_PORT))
    sock.send(b'\000\n')
    sock.close()

def format_date(ts):
    return time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(ts))


def format_record(record):
    record['date'] = format_date(record['created'])
    record['message'] = '   '.join(record['msg'].splitlines())
    return (
        u'%(date)s %(levelname)s %(name)s: (%(filename)s:%(lineno)s %(funcName)s)    %(message)s'
        % record
    )


def format_log(filename):
    """Format a given fussy log server file into more traditional format"""
    for line in open(filename):
        try:
            if line.strip():
                print((format_record(json.loads(line.strip())).encode('utf-8')))
        except Exception:
            log.warning('Failed to process line: %r', line)


def format_log_main():
    if sys.argv[1:]:
        for filename in sys.argv[1:]:
            format_log(filename)
        return 0
    else:
        print('fussy-format-log filename [filename...]')
        return 1


if __name__ == "__main__":
    main()



import asyncio, logging
from . import inotify

log = logging.getLogger(__name__)


def watch_inotify(fd: inotify.Inotify) -> asyncio.Queue:
    """Asynchronously watch the given Inotify pushing events to returned queue
    
    notify = inotify.Inotify()
    notify.add_watch(...)
    queue = watch_inotify(notify)
    while True:
        evt = await queue.get()
        if process(evt):
            notify.close()
            break
    """
    loop = asyncio.get_event_loop()
    queue = asyncio.Queue()

    def on_ready():
        if fd.closed:
            loop.remove_reader(fd)
            return
        for evt in fd.get_events(block=False):
            queue.put_nowait(evt)

    asyncio.get_event_loop().add_reader(fd, on_ready)
    return queue



"""Async wrapper around Inotify to provide event-iterable"""
from __future__ import absolute_import
from fussy import inotify
import asyncio


async def iter_events(watch):
    queue = asyncio.Queue()

    async def push_to_queue():
        for event in inotify.get_events_nonblocking(watch):
            event.fd = watch
            event.wd = watch.watches.get(event.wd, event.wd)
            await queue.put(event)

    def on_readable():
        asyncio.ensure_future(push_to_queue())

    reader = asyncio.get_event_loop().add_reader(int(watch), on_readable)

    event = True
    while not watch.closed:
        event = await queue.get()
        if event is None:
            break
        yield event


def main():
    import sys

    target = sys.argv[1]
    watch = inotify.Inotify()
    watch.add_watch(target)

    async def demo():
        print("Touch or otherwise modify first argument to produce 5 events")
        count = 0
        async for event in iter_events(watch):
            print("Event", event)
            count += 1
            if count > 4:
                break

    asyncio.get_event_loop().run_until_complete(demo())


if __name__ == "__main__":
    main()



"""JSONLogServer client for Python logging

To install a JSON Log Server client (logger), do the following::

    from fussy.jsonlog import JSONSocketHandler
    
    root = logging.getLogger( '' )
    handler = JSONSocketHandler( )
    # important, the logger is *not* intended for high volumes of 
    # messages, it is intended for messages you might actually want to 
    # look at, so set the level high!
    handler.setLevel( logging.WARN )
    root.addHandler( handler )

To add to a Django logging config::

    'handlers': {
        'json': {
            'level': 'WARN',
            'class': 'fussy.jsonlog.JSONSocketHandler',
        },
    },
    'loggers': {
        '': {
            'handlers': ['json'],
        }
    }

Records stored look like this:

    {
      "name": "root", # logger name
      "created": 1354250362.095767, # unix timestamp (time.time())
      "args": [], # unicode(a) for a in args
      "module": "<stdin>", # python module from which the message comes
      "funcName": "<module>", # function from which the message was generated
      "levelno": 30, # numeric code for the level
      "pathname": "<stdin>", # path to the file/module
      "lineno": 1, # line at which the logging message is produced
      "msg": "Hello world", # message, with *no* substitutions performed!
      "filename": "<stdin>", # the file name for the code generating the error
      "levelname": "WARNING" # the textual name of the level
    }
    
    
"""
import logging
import logging.handlers
import socket
import json
import time
import traceback
import threading
import subprocess
import os

try:
    from configparser import ConfigParser
except ImportError:
    from ConfigParser import SafeConfigParser as ConfigParser

LOGGING_PORT = 2514
try:
    unicode
except NameError:
    unicode = str


def as_bytes(content):
    if isinstance(content, unicode):
        return content.encode('utf-8')
    return content


def tags_from_config(configfile, segments):
    """Extract values from configfile

    configfile -- ConfigParser compatible configuration file
    segments -- list of (key, (section,subkey)) records to extract from file

    returns dictionary of tags to add to log messages
    """
    parser = ConfigParser()
    parser.read(configfile)
    result = {}
    for key, (section, subkey) in segments:
        result[key] = parser.get(section, subkey)
    return result


def default_tags(release_file='/etc/firmware/release.conf'):
    """Calculate the default device tags for logging processes"""
    tags = {}
    for key, command in [
        ('hostname', ['hostname']),
        ('username', ['whoami']),
    ]:
        try:
            output = subprocess.check_output(command).decode('utf-8').strip()
        except subprocess.CalledProcessError as err:
            output = None
        tags[key] = output
    if os.path.exists(release_file):
        tags.update(
            tags_from_config(
                release_file,
                [
                    ('product', ('release', 'product')),
                    ('version', ('release', 'revision')),
                ],
            )
        )
    return tags


class JSONSocketHandler(logging.handlers.SocketHandler):
    """Handler which sends JSON-formatted subsets of logging records to a socket

    The handler actually uses a JSONFormatter to format the records,
    and you can override the formatter by doing a ``setFormatter`` on the
    handler, should you wish to override the format logged.

    Note that messages *must* be single-line, as the log server uses newlines
    as the indication that a record may be written (so if you aren't single line
    your messages can become scrambled).
    """

    def __init__(self, host='127.0.0.1', port=LOGGING_PORT, tags=None):
        logging.handlers.SocketHandler.__init__(self, host, port)
        if not tags:
            tags = default_tags()

        if not self.formatter:
            self.formatter = JSONFormatter(tags=tags)
        # we really, really, really want to get messages through, and
        # the log server should *always* be up, so be far more aggressive
        # about retrying to send messages...
        self.retryStart = 0.01
        self.retryMax = 5.0
        self.retryFactor = 2.0
        self.buffer = []
        self.lock = threading.RLock()

    def makePickle(self, record):
        """We don't want to pickle, we want to jsonify"""
        return self.formatter.format(record)

    MAX_BUFFER = 100

    def send(self, s):
        """Send our content to the other side

        This implementation stores any unsent messages locally until we can
        get to the server...
        """
        self.buffer.append(s)
        if len(self.buffer) > self.MAX_BUFFER:
            self.buffer = self.buffer[-self.MAX_BUFFER :]
        if self.sock is None:
            self.createSocket()
        if self.sock:
            while self.buffer:
                try:
                    s = self.buffer.pop(0)
                except IndexError:
                    break
                else:
                    try:
                        sentsofar = 0
                        left = len(s)
                        while left > 0 and self.sock:
                            sent = self.sock.send(s[sentsofar:])
                            sentsofar = sentsofar + sent
                            left = left - sent
                    except socket.error:
                        # s was not sent in full, so re-schedule...
                        self.buffer.insert(0, s)
                        self.sock.close()
                        self.sock = None  # so we can call createSocket next time


class JSONFormatter(logging.Formatter):
    suppress = set(
        [
            'relativeCreated',
            'process',
            'threadName',
            'thread',
            'msecs',
            'processName',
            'exc_info',
            'exc_text',
            'args',
            'message',
            'request',
        ]
    )
    MSG_FORMATTER = logging.Formatter()
    DATE_FORMAT = '%Y-%m-%d %H:%M:%S'

    def __init__(self, *args, **named):
        self.tags = named.pop('tags') if 'tags' in named else {}
        super(JSONFormatter, self).__init__(*args, **named)

    def format(self, record):
        """Format the record as a subset of fields in JSON notation

        self.MSG_FORMATTER is applied to the record to produce "msg", so that
        records do not need args to be transmitted

        Each field in self.suppress is removed from the records (mostly to
        reduce record size).

        If traceback.format_exc() returns something other than ``'None\n'`` then
        the traceback is added to the record as 'traceback' key

        The JSON is generated with sort_keys, so the format should be fairly
        consistent in the final file.
        """
        raw = record.__dict__.copy()
        # Do a basic message format...
        try:
            raw['msg'] = self.MSG_FORMATTER.format(record)
        except Exception:
            raw['msg'] = 'Unable to format message'
            pass

        if raw['levelno'] >= logging.WARNING:
            trace = traceback.format_exc()
            if trace and trace != 'None\n':
                raw['traceback'] = trace
        raw['created_formatted'] = time.strftime(
            self.DATE_FORMAT, time.localtime(raw['created'])
        )
        for key in self.suppress:
            if key in raw:
                del raw[key]
        raw.update(self.tags)
        return (
            as_bytes(
                json.dumps(raw, sort_keys=True, skipkeys=True, separators=(',', ':'))
            )
            + b'\n'
        )


def sample_log():
    handler = JSONSocketHandler()
    handler.setLevel(logging.WARNING)
    log = logging.getLogger('blue')
    log.addHandler(handler)
    log.propagate = False
    log.info('Should not show up')
    log.warning('Regular log should show up: %s', 'this')
    try:
        raise AttributeError('blue')
    except Exception:
        log.error('Hello world: %s', 'this')
        print('done')
    handler.close()


if __name__ == '__main__':
    sample_log()



#! /usr/bin/env python
"""Install a given firmware package onto the file system (commands fussy-install and fussy-clean)

Major TODO items:

* TODO: error handling for all of the big issues (disk-space, memory, script failures)
"""
import subprocess, os, shutil, glob, logging, traceback
from optparse import OptionParser
from fussy import unpack, errors, nbio

log = logging.getLogger(__name__)
from ._shims import bytes, unicode

CURRENT_LINK = 'current'
FAILSAFE_NAME = 'failsafe'
PROTECTED = [CURRENT_LINK, FAILSAFE_NAME]
DEFAULT_FIRMWARE_DIRECTORY = '/opt/firmware'
LOG_FORMAT = '%(levelname)- 7s %(name)-30s %(asctime)s  %(message)s'


def clean(target=DEFAULT_FIRMWARE_DIRECTORY, protected=None):
    """Naive cleaning implementation

    Removes all names in target which are not in protected,
    paths in protected must be *full* path names, as returned
    by glob.  The target of the current link is protected.
    """
    if protected is None:
        protected = [os.path.join(target, p) for p in PROTECTED]
    # current target is always protected...
    target = final_path(target)
    current = os.path.join(target, CURRENT_LINK)
    assert os.path.exists(
        current
    ), "Current link appears to be missing, corrupt installation  (e.g. run install)?"
    current_target = final_path(current)
    assert os.path.exists(
        current_target
    ), "Current link appears to be broken, fix before cleaning (e.g. run install)"
    protected.append(current_target)
    for path in glob.glob(os.path.join(target, '*')):
        if path not in protected:
            log.warning('Removing unused firmware: %s', path)
            shutil.rmtree(path, True)
    return current_target


def final_path(link):
    """Get the final path of the given link

    raises IOError if link target does not exist, or the target is not a directory

    returns normalized real target path of the link
    """
    real = os.path.normpath(os.path.realpath(link))
    if not os.path.exists(real):
        raise IOError("Target of link %(link)r (%(real)s) does not exist" % locals())
    if not os.path.isdir(real):
        raise IOError("Target of link %(link)r is not a directory" % locals())
    return real


def swap_link(final_target, current):
    """Swap current link to point to final_target

    Steps taken:

        * if there is an existing tmp link, remove it
        * create a tmp link to the final target
        * rename tmp link to `current`

    returns None
    """
    tmp = current + '~'
    try:
        os.remove(tmp)
        # TODO: address race condition, you should have the while upgrade
        # cron-locked, but that's not obvious here...
    except (OSError, IOError):
        pass
    os.symlink(os.path.basename(final_target), tmp)
    os.rename(tmp, current)


def install_bytes(filename, keyring='/etc/fussy/keys', target='/opt/firmware'):
    """Install the packaged bytes into a final target directory

    Steps taken:

        * unpack firmware using :func:`fussy.unpack.unpack`
        * rsync new_firmware into /opt/firmware (`target`)
            * if `CURRENT_LINK` (current) is present in `target`,
              will hard-link shared files between the new firmware and `current`
              to reduce disk use (using :command:`rsync` parameter --link-dest)
        * removes the temporary directory where unpacking was performed

    returns full path to sub-directory of target where new firmware was installed

    raises Errors on most failures, including disk-full, failed commands, missing
    executables, etc
    """
    temp_dir = unpack.unpack(filename, keyring)
    assert os.path.exists(temp_dir)
    base_name = os.path.basename(temp_dir)
    try:
        final_target = os.path.join(os.path.normpath(target), base_name)
        i = 0
        while os.path.exists(final_target):
            i += 1
            final_target = os.path.join(
                os.path.normpath(target), base_name + '-%i' % (i,)
            )
        current = os.path.join(os.path.normpath(target), 'current')
        command = ['rsync', '-aq']
        if os.path.exists(current):
            log.info('Reducing firmware size with hard-link compression')
            command.append('--link-dest=%s' % (current,))
        # TODO: figure out some way to configure rsync to not create a second
        # level directory when told `rsync -a a b`
        command.extend([os.path.join(temp_dir, x) for x in os.listdir(temp_dir)])
        command.extend([final_target])
        log.info('Fixating firmware')
        subprocess.check_call(command)
        return final_target
    finally:
        shutil.rmtree(temp_dir, True)


def enable(final_target, current, recovery=False):
    """Attempt to enable final_target as the current release

    Steps taken:

        * runs `final_target/.pre-install final_target recovery`
          (iff .pre-install is present)
        * (atomically) swaps the link `current`
          for a link that points to `final_target`
        * runs `final_target/.post-install final_target recovery`
          (iff .post-install is present)
        * if a failure occurs before swap-link completes,
          deletes final_target
        * if a failure occurs after swap-link, raises an error

    returns None
    raises Exceptions on lots of failure cases
    """
    pre_install = os.path.join(final_target, '.pre-install')
    post_install = os.path.join(final_target, '.post-install')

    try:
        if os.path.exists(pre_install):
            log.info('Running pre-install script')

            def report_progress(line):
                if isinstance(line, bytes):
                    try:
                        line = line.decode('utf-8')
                    except UnicodeDecodeError as err:
                        line = line.decode('latin-1')
                log.info('pre-install: %s', line)

            pipe = (
                nbio.Process(
                    [pre_install, final_target, unicode(recovery).encode('utf-8')],
                    by_line=True,
                    stderr=-1,
                )
                | report_progress
            )
            pipe()
        log.info('Setting firmware current')
        swap_link(final_target, current)
    except Exception:
        # we failed in either pre-setup or swapping
        if not recovery:
            log.warning('Failed during pre-install or swap, aborting')
            shutil.rmtree(final_target, True)
            raise
        else:
            log.exception(
                'Failure attempting to re-initialize previous installation, likely reboot required'
            )
            # Note, we will attempt to run post-install, as it is often able to perform
            # cleanups that the pre-install cannot...

    if os.path.exists(post_install):
        log.info('Running post-install script')

        def report_progress(line):
            log.info('post-install: %s', line)

        pipe = (
            nbio.Process(
                [post_install, final_target, unicode(recovery).encode('utf-8')],
                by_line=True,
                stderr=-1,
            )
            | report_progress
        )
        pipe()


def ensure_current_link(current, failsafe):
    """Ensure that current is a link (not a directory)"""
    if os.path.isdir(current) and not os.path.islink(current):
        if not os.path.exists(failsafe):
            # this is the initial install case, where we want failsafe to be
            # created from the initial installation image...
            os.rename(current, failsafe)
            os.symlink(os.path.basename(failsafe), current)
        else:
            raise RuntimeError(
                """%(current)r is a directory, it should be a link.  Failsafe already exists."""
                % locals()
            )


def install(
    filename, keyring='/etc/fussy/keys', target='/opt/firmware', unpacked=False
):
    """Install given firmware <filename> into given target directory

    Steps taken:

        * unpack firmware (using :func:`fussy.install.install_bytes`)
        * enable firmware (using :func:`fussy.install.enable`)
        * if :func:enable fails, enable `previous`
          (or `failsafe` if there was no previous)

    returns (error_code (0 is success), path name of the installed package)
    """
    current = os.path.join(os.path.normpath(target), CURRENT_LINK)
    failsafe = os.path.join(os.path.normpath(target), FAILSAFE_NAME)

    ensure_current_link(current, failsafe)

    previous = None
    try:
        previous = final_path(current)
    except IOError as err:
        log.warning("Target of current does not appear to exist: %s", err)
    if not previous and os.path.exists(failsafe):
        previous = failsafe
    log.info('Previous installation: %s', previous)
    log.info('Unpacking firmware to disk')
    if not unpacked:
        final_target = install_bytes(filename, keyring, target)
    else:
        final_target = os.path.normpath(os.path.abspath(filename))
        if not final_target.startswith(target):
            raise RuntimeError("Firmware is not unpacked into the target directory")
        if os.path.basename(final_target) == 'current':
            raise RuntimeError("Firmware is named 'current', cannot activate")
    log.info('New installation: %s', final_target)
    assert os.path.exists(final_target)
    try:
        enable(final_target, current)
    except Exception as err:
        log.error("Failure installing %s: %s", final_target, err)
        log.error("Traceback: %s", traceback.format_exc())
        if previous:
            log.warning("Attempting to restore previous: %s", previous)
            enable(previous, current, recovery=True)
            raise errors.RevertedFailure(previous)
        log.error("Unable to recover, contact support!")
        raise errors.UnrecoverableError(unicode(err))
        # TODO: need lots more logic in the back-off code...
    else:
        log.info("Successfully installed %s", final_target)
        return final_target


def get_options():
    """Creates the OptionParser used in :func:`main` """
    parser = OptionParser()
    parser.add_option(
        '-f',
        '--file',
        dest='file',
        default=None,
        action="store",
        type="string",
        help="The firmware archive to unpack, must be a .tar.gz.gpg or a .tar.gz.asc",
    )
    parser.add_option(
        '-k',
        '--keyring',
        dest='keyring',
        default=unpack.DEFAULT_KEYRING,
        action="store",
        type="string",
        help="GPG keyring to use for verification/decryption (default /etc/fussy/keys)",
    )
    parser.add_option(
        '-t',
        '--target',
        dest='target',
        default=DEFAULT_FIRMWARE_DIRECTORY,
        action="store",
        type="string",
        help="Directory into which to rsync the firmware (default /opt/firmware)",
    )
    parser.add_option(
        '-u',
        '--unpacked',
        dest='unpacked',
        default=False,
        action="store_true",
        help="Target is pre-unpacked image directory (fussy-install-bytes)",
    )
    parser.add_option(
        '-l',
        '--logfile',
        dest='logfile',
        default='/tmp/fussy-install.log',
        help="File into which to write the fussy installation log (default /tmp/fussy-install.log)",
    )
    return parser


def configure_log(logfile):
    """Configure logging to the given log-file

    You can use this function in your `.post-install` scripts
    to send your logs to the given log-file in the same format
    as the logs fussy itself produces.
    """
    if logfile:
        logging.basicConfig(
            filename=logfile,
            level=logging.INFO,
            format=LOG_FORMAT,
            datefmt='%Y-%m-%d %H:%M:%S',
        )
    else:
        logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)
    return logfile


def main():
    """Main entry-point for the fussy-install script

    Steps taken:

        * parses arguments
        * launches :func:`install`
    """
    parser = get_options()
    options, args = parser.parse_args()
    if not options.file:
        if args:
            options.file = args[0]
        else:
            parser.error("Need a file to install")
    configure_log(options.logfile)
    try:
        install(
            options.file, options.keyring, options.target, unpacked=options.unpacked
        )
        return 0
    except Exception as err:
        log.error("Failure during installation: %s", err)
        log.error("Traceback: %s", traceback.format_exc())
        raise


def clean_main():
    """Main entry-point for fussy-clean script

    Steps taken:

        * parses arguments
        * launches :func:`clean`
    """
    parser = OptionParser()
    parser.add_option(
        '-t',
        '--target',
        dest='target',
        default=DEFAULT_FIRMWARE_DIRECTORY,
        action="store",
        type="string",
        help="Directory into which to rsync the firmware (default /opt/firmware)",
    )
    parser.add_option(
        '-l',
        '--logfile',
        dest='logfile',
        default='/tmp/fussy-install.log',
        help="File into which to write the fussy installation log (default /tmp/fussy-install.log)",
    )
    options, args = parser.parse_args()
    configure_log(options.logfile)
    clean(options.target)
    return 0


def install_bytes_main_options():
    """Creates the OptionParser used in :func:`unpack_main` """
    parser = OptionParser("Unpacks a firmware onto disk without activating it")
    parser.add_option(
        '-f',
        '--file',
        dest='file',
        default=None,
        action="store",
        type="string",
        help="The firmware archive to unpack, must be a .tar.gz.gpg or a .tar.gz.asc",
    )
    parser.add_option(
        '-k',
        '--keyring',
        dest='keyring',
        default=unpack.DEFAULT_KEYRING,
        action="store",
        type="string",
        help="GPG keyring to use for verification/decryption (default /etc/fussy/keys)",
    )
    parser.add_option(
        '-t',
        '--target',
        dest='target',
        default=DEFAULT_FIRMWARE_DIRECTORY,
        action="store",
        type="string",
        help="Directory into which to rsync the firmware (default /opt/firmware)",
    )
    parser.add_option(
        '-l',
        '--logfile',
        dest='logfile',
        default='/tmp/fussy-install.log',
        help="File into which to write the fussy installation log (default /tmp/fussy-install.log)",
    )
    return parser


def install_bytes_main():
    parser = install_bytes_main_options()
    options, args = parser.parse_args()
    if not options.file:
        if args:
            options.file = args[0]
        else:
            parser.error("Need a file to install")
    configure_log(options.logfile)
    try:
        final_target = install_bytes(options.file, options.keyring, options.target)
        print(final_target)
        return 0
    except Exception as err:
        log.error("Failure during unpacking: %s", err)
        log.error("Traceback: %s", traceback.format_exc())
        raise



#! /usr/bin/env python
"""Bundle a package as a signed firmware for installation/redistribution
"""
import tempfile, subprocess, os
from optparse import OptionParser

DEFAULT_EXCLUDES = ['.svn', '.bzr', '.git']


def pack(root_dir, excludes=None, tar_name=None):
    """Bundle directory into a firmware file...
    
    * root_dir -- directory to be packed into the firmware, the 
        :func:`os.path.basename` of this directory will be the name used 
        in the created bundle filename and the directory installed into 
        the target on client machines
    * excludes -- patterns to exclude from the firmware image (passed to
        :command:`tar` during packing)
    
    returns absolute filename for generated .tar.gz firmware 
    (created in a temporary directory)
    """
    temp_dir = tempfile.mkdtemp(prefix='fussy-', suffix='-pack')
    firmware_name = os.path.basename(root_dir)
    if firmware_name == 'current':
        raise RuntimeError('Cannot create a firmware named "current"')
    if tar_name is None:
        tar_name = firmware_name + '.tar.gz'
    tar_file = os.path.join(temp_dir, tar_name)
    command = (
        ['tar', '-zcf', tar_file]
        + ['--exclude=%s' % (x,) for x in ((excludes or []) + DEFAULT_EXCLUDES)]
        + [os.path.basename(root_dir)]
    )
    subprocess.check_call(command, cwd=os.path.dirname(os.path.abspath(root_dir)))
    return tar_file


def sign(tar_file, key=None, encrypt_for=None):
    """Sign an existing tar.gz-file with the default key
    
    * tar_file -- tar.gz file to sign (and/or encrypt)
    * encrypt_for -- if non-null, string identifier for which to encrypt
    """
    gpg_file = tar_file + '.gpg'
    command = ['gpg']
    if encrypt_for:
        command += ['-se', '-r', encrypt_for]
    else:
        command += ['--sign']
    if key:
        command += ['-u', key]
    command.append(tar_file)
    subprocess.check_call(command)
    assert os.path.exists(gpg_file)
    os.remove(tar_file)
    return gpg_file


def get_options():
    """Produces the OptionParser for :func:main"""
    parser = OptionParser()
    parser.add_option(
        '-x',
        '--exclude',
        dest='exclude',
        action="append",
        type="string",
        help="Paths/patterns to exclude from the archive",
    )
    parser.add_option(
        '-r',
        '--root',
        dest='root',
        default='firmware',
        type="string",
        help="The directory to be packed for distribution, should be a versioned/unique name, for instance including a human-readable timestamp",
    )
    parser.add_option(
        '-u',
        '--unsigned',
        dest='signed',
        action="store_false",
        default=True,
        help="If specified, do not do signing, just produced an unsigned tar.gz file (intended to allow for signing on a separate workstation)",
    )
    parser.add_option(
        '-t',
        '--tarname',
        dest='tar_name',
        default=None,
        type="string",
        help="If specified, override the calculation of the .tar.gz filename (defaults to target directory basename + .tar.gz)",
    )
    add_encryption_options(parser)
    return parser


def add_encryption_options(parser):
    parser.add_option(
        '-k',
        '--key',
        dest='key',
        default=None,
        help='If specified, use this key id to perform the signing (passed to :cmd:gpg -u option), otherwise uses default identity',
    )
    parser.add_option(
        '-e',
        '--encrypt-for',
        dest='encrypt',
        action="store",
        type="string",
        help="The name of the key for which to encrypt (otherwise just sign)",
    )
    return parser


def main():
    """Main function for the packing script"""
    options, args = get_options().parse_args()
    tar_file = pack(options.root, options.exclude, options.tar_name)
    if options.signed:
        gpg_file = sign(tar_file, key=options.key, encrypt_for=options.encrypt)
        print(gpg_file)
    else:
        print(tar_file)
    return 0


def get_sign_options():
    parser = OptionParser()
    parser.add_option(
        '-f',
        '--filename',
        dest='filename',
        default=None,
        help='The full path to the tar.gz file we will sign',
    )
    add_encryption_options(parser)
    return parser


def sign_main():
    """Main function that *just* signs a tar-file"""
    options, args = get_sign_options().parse_args()
    if not options.filename and args:
        options.filename = args[0]
    gpg_file = sign(options.filename, key=options.key, encrypt_for=options.encrypt)
    print(gpg_file)
    return 0



"""Transactional write/commit for config files"""
from . import twrite, nbio
import os, logging, tempfile, shutil

log = logging.getLogger(__name__)


def commonpath(paths):
    if hasattr(os.path, 'commonpath'):
        return os.path.commonpath(paths)
    else:
        abs_paths = [os.path.normpath(os.path.abspath(p)).split('/') for p in paths]
        result = []
        for segments in zip(*abs_paths):
            element = segments[0]
            for seg in segments[1:]:
                if seg != element:
                    return '/'.join(result) if result else '/'
            result.append(element)
        return '/'.join(result)


class FSTransaction(object):
    def __init__(self, target_dir, work_dir=None):
        self.target_dir = target_dir.rstrip('/')
        self.temp_path = work_dir or os.path.join(self.target_dir, '.transaction')
        self.writer = os.path.join(self.temp_path, 'write')
        self.recover = os.path.join(self.temp_path, 'recover')
        if not os.path.exists(self.writer):
            os.makedirs(self.writer)
        if not os.path.exists(self.recover):
            os.makedirs(self.recover)
        self.to_delete = []
        self.recover_delete = []
        self.in_transaction = []
        self.in_recovery = []

    def copy(self, source, path):
        """Copy the given source to the target path"""
        relative = self.relative_path(path)
        if os.path.exists(source):
            return self._copy_source_to_root(source, self.writer, relative)
        return None

    def _copy_source_to_root(self, source, root, relative):
        """Copy source-file into root/relative"""
        recovery = os.path.join(root, relative)
        dirname = os.path.dirname(recovery)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        if os.path.islink(source):
            link_dest = os.readlink(source)
            if os.path.exists(recovery):
                os.remove(recovery)
            os.symlink(link_dest, recovery)
        else:
            shutil.copy2(source, os.path.join(root, relative))
        return recovery

    def twrite(self, path, content):
        """Transactional twrite operation"""
        relative = self.save_path(path)
        if not os.path.exists(path):
            self.recover_delete.append(path)
        written = os.path.join(self.writer, relative.lstrip('/'))
        twrite.twrite(written, content)
        return written

    def chmod(self, path, mode=0o644):
        """Set mode of the path to given mode within transaction"""
        relative = self.save_path(path)
        written = os.path.join(self.writer, relative)
        if not os.path.exists(written):
            shutil.copy2(path, written)
        os.chmod(written, mode)

    def chown(self, path, *args, **named):
        """Change ownership of a file within transaction"""
        relative = self.save_path(path)
        written = os.path.join(self.writer, relative)
        if not os.path.exists(written):
            shutil.copy2(path, written)
        os.chown(written, *args, **named)

    def relative_path(self, path):
        if not path.startswith('/'):
            path = os.path.join(self.target_dir, path)
        prefix = commonpath([self.target_dir, path])
        if prefix != self.target_dir:
            raise ValueError("Need a sub-path of %r: %r", self.target_dir, path)
        relative = path[len(prefix) :].lstrip('/')
        return relative

    def writer_path(self, path):
        """Calculate the path for writing to given path"""
        return os.path.join(self.writer, self.relative_path(path))

    def recovery_path(self, path):
        """Calculate the path for saving files for restore"""
        return os.path.join(self.writer, self.relative_path(path))

    def debug_description(self):
        """Create a debugging description of our operations"""
        description = nbio.Process(
            [
                'tree',
                self.temp_path,
            ]
        )().decode('utf-8', 'ignore')
        return description

    def save_path(self, path):
        """Utility to save the content, state, ownership etc of file"""
        relative = self.relative_path(path)
        if os.path.exists(path):
            self._copy_source_to_root(path, self.recover, relative)
        return relative

    def remove(self, path):
        """Delete a path during the transaction"""
        relative = self.save_path(path)
        self.to_delete.append(os.path.join(self.target_dir, relative))

    def symlink(self, target, link):
        """Create link pointing to target"""
        relative = self.save_path(link)
        final = os.path.join(self.writer, relative)
        os.symlink(target, final)
        return final

    def add_to_transaction(self, command, *args, **named):
        """Add a callable that must run inside the transaction (failure cancels)"""
        self.in_transaction.append((command, args, named))

    def add_to_recovery(self, command, *args, **named):
        """Add a callable that should run after recovery"""
        self.in_recovery.append((command, args, named))

    def would_write(self, path):
        """Would we write to the given path?"""
        relative = self.relative_path(path)
        return os.path.exists(os.path.join(self.writer, relative))

    def would_remove(self, path):
        """Would we delete the given path?"""
        relative = self.relative_path(path)
        return os.path.join(self.target_dir, relative) in self.to_delete

    def _rsync_source_to_dest(self, source, dest):
        """Run rsync from source to destination"""
        command = (
            ['rsync', '-av']
            + [os.path.join(source, x) for x in os.listdir(source)]
            + [
                dest + '/',
            ]
        )
        log.info(" %s", " ".join(command))
        try:
            nbio.Process(command)()
        except nbio.ProcessError as err:
            log.warning("rsync operation failed")
            raise RuntimeError('Failure during %s' % (" ".join(command)))
        else:
            return True

    def commit(self):
        """Attempt to commit changes to the transaction"""
        try:
            self._rsync_source_to_dest(self.writer, self.target_dir)
            for filename in self.to_delete:
                log.debug("  rm: %s", filename)
                if os.path.exists(filename):
                    os.remove(filename)
            for callable, args, named in self.in_transaction:
                log.debug("  run: %s", callable)
                callable(*args, **named)
        except Exception as err:
            log.exception("Unable to commit changes")
            try:
                self.abort()
            except Exception as abort_err:
                log.exception("Failure during abort, filesystem state inconsistent")
                raise abort_err
            return False
        else:
            self.cleanup()
            return True

    def abort(self):
        """Abort commit, attempt to restore previous"""
        log.debug("Aborting from: %s", self.target_dir)
        self._rsync_source_to_dest(self.recover, self.target_dir)
        for filename in self.recover_delete:
            if os.path.exists(filename):
                os.remove(filename)
        for callable, args, named in self.in_recovery:
            callable(*args, **named)

    def cleanup(self):
        shutil.rmtree(self.temp_path)